#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>

#include "ndalloc.h"

#define N 101

#define CUBE(x) ((x)*(x)*(x))

static double wtime (void) {
  struct timeval now;

  (void) gettimeofday(&now, NULL);

  return now.tv_sec + 1e-6*now.tv_usec;
}

int main (void) {
  size_t i, j, k;
  size_t counter;
  double timer;

  size_t n;
  double h;
  double kappa;
  double dt;
  double alpha;
  double T;

  double t;
  double *** u;
  double *** u_prev;

  n = N;
  h = 1.0/n;
  dt = 1e-3;
  kappa = 1.0;

  alpha = dt*kappa / (h*h);

  T = 2.0;

  u = ndalloc(3, sizeof(double), n, n, n);
  u_prev = ndalloc(3, sizeof(double), n, n, n);

  if (u == NULL || u_prev == NULL) {
    perror("ndalloc");
    exit(EXIT_FAILURE);
  }

  /* set everything to zero (this includes the boundary) */
  memset(&u_prev[0][0][0], 0, n*n*n * sizeof(double));

  /* interior initial conditions */
  for (i = 1; i < n-1; i++)
    for (j = 1; j < n-1; j++)
      for (k = 1; k < n-1; k++)
	u_prev[i][j][k] = 1.0;

  /* top boundary condition */
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      u_prev[i][j][n-1] = 1.0;

  /* ensure that the two arrays are identical */
  memcpy(&u[0][0][0], &u_prev[0][0][0],
	 n*n*n * sizeof(double));

  t = 0.0;
  counter = 0;

  timer = wtime();
  while (t < T) {
    /* compute next level */
    for (i = 1; i < n-1; i++)
      for (j = 1; j < n-1; j++)
	for (k = 1; k < n-1; k++)
	  u[i][j][k] = u_prev[i][j][k] +
	    alpha * (u_prev[i+1][j][k] + u_prev[i-1][j][k] +
		     u_prev[i][j+1][k] + u_prev[i][j-1][k] +
		     u_prev[i][j][k+1] + u_prev[i][j][k-1] -
		     6 * u_prev[i][j][k]);

    /* swap pointers */
    {
      double *** temp = u;
      u = u_prev;
      u_prev = temp;
    }

    counter += 1;
    t += dt;
  }
  timer = wtime() - timer;

  /*
   * FLOPS:
   *
   * Per inner iteration,
   * + : 6 times
   * - : 1 time
   * * : 2 times
   * total of 9 operations.
   *
   * Inner iterations per outer iteration:
   * (n-2) * (n-2) * (n-2) = (n-2)^3.
   *
   * Total operations:
   * 9*(n-2)^3*counter.
   */
  printf("GFLOPS: %f\n", 1e-9*9*CUBE(n-2)*counter/timer);

  ndalloc_free(3, u);
  ndalloc_free(3, u_prev);

  return 0;
}
